searchNodes=[{"ref":"Xandra.html","title":"Xandra","type":"module","doc":"This module provides the main API to interface with Cassandra. This module handles the connection to Cassandra, queries, connection pooling, connection backoff, logging, and more. Many of these features are provided by the DBConnection library, which Xandra is built on top of. Errors Many of the functions in this module (whose names don&#39;t end with a !) return values in the form {:ok, result} or {:error, error}. While result varies based on the specific function, error is always one of the following: a Xandra.Error struct: such structs represent errors returned by Cassandra. When such an error is returned, it means that communicating with the Cassandra server was successful, but the server returned an error. Examples of these errors are syntax errors in queries, non-existent tables, and so on. See Xandra.Error for more information. a Xandra.ConnectionError struct: such structs represent errors in the communication with the Cassandra server. For example, if the Cassandra server dies while the connection is waiting for a response from the server, a Xandra.ConnectionError error will be returned. See Xandra.ConnectionError for more information. Parameters, encoding, and types Xandra supports parameterized queries (queries that specify &quot;parameter&quot; values through ? or :named_value): SELECT * FROM users WHERE name = ? AND email = ? SELECT * FROM users WHERE name = :name AND email = :email When a query has positional parameters, parameters can be passed as a list to functions like execute/4: in this case, a parameter in a given position in the list will be used as the ? in the corresponding position in the query. When a query has named parameters, parameters are passed as a map with string keys representing each parameter&#39;s name and values representing the corresponding parameter&#39;s value. Types Cassandra supports many types of values, and some types have &quot;shades&quot; that cannot be represented by Elixir types. For example, in Cassandra an integer could be a &quot;bigint&quot; (a 64 bit integer), an &quot;int&quot; (a 32 bit integer), a &quot;smallint&quot; (a 16 bit integer), or others; in Elixir, however, integers are just integers (with varying size to be precise), so it is impossible to univocally map Elixir integers to a specific Cassandra integer type. For this reason, when executing simple parameterized queries (statements) it is necessary to explicitly specify the type of each value. To specify the type of a value, that value needs to be provided as a two-element tuple where the first element is the value&#39;s type and the second element is the value itself. Types are expressed with the same syntax used in CQL: for example, 16-bit integers are represented as &quot;smallint&quot;, while maps of strings to booleans are represented as &quot;map&lt;text, boolean&gt;&quot;. # Using a list of parameters: statement = &quot;INSERT INTO species (name, properties) VALUES (?, ?)&quot; Xandra.execute(conn, statement, [ {&quot;text&quot;, &quot;human&quot;}, {&quot;map&lt;text, boolean&gt;&quot;, %{&quot;legs&quot; =&gt; true, &quot;arms&quot; =&gt; true, &quot;tail&quot; =&gt; false}}, ]) # Using a map of parameters: statement = &quot;INSERT INTO species (name, properties) VALUES (:name, :properties)&quot; Xandra.execute(conn, statement, %{ &quot;name&quot; =&gt; {&quot;text&quot;, &quot;human&quot;}, &quot;properties&quot; =&gt; {&quot;map&lt;text, boolean&gt;&quot;, %{&quot;legs&quot; =&gt; true, &quot;arms&quot; =&gt; true, &quot;tail&quot; =&gt; false}}, }) You only need to specify types for simple queries (statements): when using prepared queries, the type information of each parameter of the query is encoded in the prepared query itself. # Using a map of parameters: prepared = Xandra.prepare!(conn, &quot;INSERT INTO species (name, properties) VALUES (:name, :properties)&quot;) Xandra.execute(conn, prepared, %{ &quot;name&quot; =&gt; &quot;human&quot;, &quot;properties&quot; =&gt; %{&quot;legs&quot; =&gt; true, &quot;arms&quot; =&gt; true, &quot;tail&quot; =&gt; false}, }) User-defined types Xandra supports user-defined types (UDTs). A UDT can be inserted as a map with string fields. For example, consider having created the following UDTs: CREATE TYPE full_name (first_name text, last_name text) CREATE TYPE profile (username text, full_name frozen&lt;full_name&gt;) and having the following table: CREATE TABLE users (id int PRIMARY KEY, profile frozen&lt;profile&gt;) Inserting rows will look something like this: prepared_insert = Xandra.prepare!(conn, &quot;INSERT INTO users (id, profile) VALUES (?, ?)&quot;) profile = %{ &quot;username&quot; =&gt; &quot;bperry&quot;, &quot;full_name&quot; =&gt; %{&quot;first_name&quot; =&gt; &quot;Britta&quot;, &quot;last_name&quot; =&gt; &quot;Perry&quot;}, } Xandra.execute!(conn, prepared_insert, [_id = 1, profile]) Note that inserting UDTs is only supported on prepared queries. When retrieved, UDTs are once again represented as maps with string keys. Retrieving the row inserted above would look like this: %{&quot;profile&quot; =&gt; profile} = conn |&gt; Xandra.execute!(&quot;SELECT id, profile FROM users&quot;) |&gt; Enum.fetch!(0) profile #=&gt; %{&quot;username&quot; =&gt; &quot;bperry&quot;, &quot;full_name&quot; =&gt; %{&quot;first_name&quot; =&gt; &quot;Britta&quot;, &quot;last_name&quot; =&gt; &quot;Perry&quot;}} Reconnections Thanks to the DBConnection library, Xandra is able to handle connection losses and to automatically reconnect to Cassandra. By default, reconnections are retried at exponentially increasing randomized intervals, but backoff can be configured through a subset of the options accepted by start_link/2. These options are described in the documentation for DBConnection.start_link/2. Clustering Xandra supports connecting to multiple nodes in a Cassandra cluster and executing queries on different nodes based on load balancing strategies. See the documentation for Xandra.Cluster for more information. Authentication Xandra supports Cassandra authentication. See the documentation for Xandra.Authenticator for more information. Retrying failed queries Xandra takes a customizable and extensible approach to retrying failed queries through &quot;retry strategies&quot; that encapsulate the logic for retrying queries. See Xandra.RetryStrategy for documentation on retry strategies. Compression Xandra supports compression. To inform the Cassandra server that the connections you start should use compression for data transmitted to and from the server, you can pass the :compressor option to start_link/1; this option should be a module that implements the Xandra.Compressor behaviour. After this, all compressed data that Cassandra sends to the connection will be decompressed using this behaviour module. To compress outgoing data (such as when issuing or preparing queries), the :compressor option should be specified explicitly. When it&#39;s specified, the given module will be used to compress data. If no :compressor option is passed, the outgoing data will not be compressed."},{"ref":"Xandra.html#execute/3","title":"Xandra.execute/3","type":"function","doc":"Executes the given simple query, prepared query, or batch query. Returns {:ok, result} if executing the query was successful, or {:error, error} otherwise. The meaning of the params_or_options argument depends on what query is: if query is a batch query, than params_or_options has to be a list of options that will be used to run the batch query (since batch queries don&#39;t use parameters as parameters are attached to each query in the batch). if query is a simple query (a string) or a prepared query, then params_or_opts is a list or map of parameters, and this function is exactly the same as calling execute(conn, query, params_or_options, []). When query is a batch query, successful results will always be Xandra.Void structs. When {:error, error} is returned, error can be either a Xandra.Error or a Xandra.ConnectionError struct. See the module documentation for more information on errors. Options for batch queries When query is a batch query, params_or_options is a list of options. All options supported by DBConnection.execute/4 are supported, and the following additional batch-specific options: :consistency - same as the :consistency option described in the documentation for execute/4. :serial_consistency - same as the :serial_consistency option described in the documentation for execute/4. :timestamp - (integer) using this option means that the provided timestamp will apply to all the statements in the batch that do not explicitly specify a timestamp. Examples For examples on executing simple queries or prepared queries, see the documentation for execute/4. Examples below specifically refer to batch queries. See the documentation for Xandra.Batch for more information about batch queries and how to construct them. prepared_insert = Xandra.prepare!(conn, &quot;INSERT (email, name) INTO users VALUES (?, ?)&quot;) batch = Xandra.Batch.new() |&gt; Xandra.Batch.add(prepared_insert, [&quot;abed@community.com&quot;, &quot;Abed Nadir&quot;]) |&gt; Xandra.Batch.add(prepared_insert, [&quot;troy@community.com&quot;, &quot;Troy Barnes&quot;]) |&gt; Xandra.Batch.add(prepared_insert, [&quot;britta@community.com&quot;, &quot;Britta Perry&quot;]) # Execute the batch: Xandra.execute(conn, batch) #=&gt; {:ok, %Xandra.Void{}} # Execute the batch with a default timestamp for all statements: Xandra.execute(conn, batch, timestamp: System.system_time(:millisecond) - 1_000) #=&gt; {:ok, %Xandra.Void{}} All DBConnection.execute/4 options are supported here as well: Xandra.execute(conn, batch, pool: DBConnection.Poolboy) #=&gt; {:ok, %Xandra.Void{}}"},{"ref":"Xandra.html#execute/4","title":"Xandra.execute/4","type":"function","doc":"Executes the given simple query or prepared query with the given parameters. Returns {:ok, result} where result is the result of executing query if the execution is successful (there are no network errors or semantic errors with the query), or {:error, error} otherwise. result can be one of the following: a Xandra.Void struct - returned for queries such as INSERT, UPDATE, or DELETE. a Xandra.SchemaChange struct - returned for queries that perform changes on the schema (such as creating tables). a Xandra.SetKeyspace struct - returned for USE queries. a Xandra.Page struct - returned for queries that return rows (such as SELECT queries). The properties of each of the results listed above are described in each result&#39;s module. Options This function accepts all options accepted by DBConnection.execute/4, plus the following ones: :consistency - (atom) specifies the consistency level for the given query. See the Cassandra documentation for more information on consistency levels. The value of this option can be one of: :one (default) :two :three :any :quorum :all :local_quorum :each_quorum :serial :local_serial :local_one :page_size - (integer) the size of a page of results. If query returns Xandra.Page struct, that struct will contain at most :page_size rows in it. Defaults to 10_000. :paging_state - (binary) the offset where rows should be returned from. By default this option is not present and paging starts from the beginning. See the &quot;Paging&quot; section below for more information on how to page queries. :timestamp - (integer) the default timestamp for the query (in microseconds). If provided, overrides the server-side assigned timestamp; however, a timestamp in the query itself will still override this timestamp. :serial_consistency - (atom) specifies the serial consistency to use for executing the given query. Can be of :serial and :local_serial. :compressor - (module) the compressor module used to compress and decompress data. See the &quot;Compression&quot; section in the module documentation. By default, this option is not present. :retry_strategy - (module) the module implementing the Xandra.RetryStrategy behaviour that is used in case the query fails to determine whether to retry it or not. See the &quot;Retrying failed queries&quot; section in the module documentation. By default, this option is not present. :date_format - (:date or :integer) controls the format in which dates are returned. When set to :integer the returned value is a number of days from the Unix epoch, a date struct otherwise. Defaults to :date. :time_format - (:time or :integer) controls the format in which times are returned. When set to :integer the returned value is a number of nanoseconds from midnight, a time struct otherwise. Defaults to :time. :timestamp_format - (:datetime or :integer) controls the format in which timestamps are returned. When set to :integer the returned value is a number of milliseconds from the Unix epoch, a datetime struct otherwise. Defaults to :datetime. Parameters The params argument specifies parameters to use when executing the query; it can be either a list of positional parameters (specified via ? in the query) or a map of named parameters (specified as :named_parameter in the query). When query is a simple query, the value of each parameter must be a two-element tuple specifying the type used to encode the value and the value itself; when query is a prepared query, this is not necessary (and values can just be values) as the type information is encoded in the prepared query. See the module documenatation for more information about query parameters, types, and encoding values. Examples Executing a simple query (which is just a string): statement = &quot;INSERT INTO users (first_name, last_name) VALUES (:first_name, :last_name)&quot; {:ok, %Xandra.Void{}} = Xandra.execute(conn, statement, %{ &quot;first_name&quot; =&gt; {&quot;text&quot;, &quot;Chandler&quot;}, &quot;last_name&quot; =&gt; {&quot;text&quot;, &quot;Bing&quot;}, }) Executing the query when atom_keys: true has been specified in Xandra.start_link/1: Xandra.execute(conn, statement, %{ first_name: {&quot;text&quot;, &quot;Chandler&quot;}, last_name: {&quot;text&quot;, &quot;Bing&quot;} }) Executing a prepared query: prepared = Xandra.prepare!(conn, &quot;INSERT INTO users (first_name, last_name) VALUES (?, ?)&quot;) {:ok, %Xandra.Void{}} = Xandra.execute(conn, prepared, [&quot;Monica&quot;, &quot;Geller&quot;]) Performing a SELECT query and using Enum.to_list/1 to convert the Xandra.Page result to a list of rows: statement = &quot;SELECT * FROM users&quot; {:ok, %Xandra.Page{} = page} = Xandra.execute(conn, statement, _params = []) Enum.to_list(page) #=&gt; [%{&quot;first_name&quot; =&gt; &quot;Chandler&quot;, &quot;last_name&quot; =&gt; &quot;Bing&quot;}, #=&gt; %{&quot;first_name&quot; =&gt; &quot;Monica&quot;, &quot;last_name&quot; =&gt; &quot;Geller&quot;}] Performing the query when atom_keys: true has been specified in Xandra.start_link/1: {:ok, page} = Xandra.execute(conn, statement, _params = []) Enum.to_list(page) #=&gt; [%{first_name: &quot;Chandler&quot;, last_name: &quot;Bing&quot;}, #=&gt; %{first_name: &quot;Monica&quot;, last_name: &quot;Geller&quot;}] Ensuring the write is written to the commit log and memtable of at least three replica nodes: statement = &quot;INSERT INTO users (first_name, last_name) VALUES (&#39;Chandler&#39;, &#39;Bing&#39;)&quot; {:ok, %Xandra.Void{}} = Xandra.execute(conn, statement, _params = [], consistency: :three) This function supports all options supported by DBConnection.execute/4; for example, if the conn connection was started with pool: DBConnection.Poolboy, then the :pool option would have to be passed here as well: statement = &quot;DELETE FROM users WHERE first_name = &#39;Chandler&#39;&quot; {:ok, %Xandra.Void{}} = Xandra.execute(conn, statement, _params = [], pool: DBConnection.Poolboy) Paging Since execute/4 supports the :paging_state option, it is possible to manually implement paging. For example, given the following prepared query: prepared = Xandra.prepare!(conn, &quot;SELECT first_name FROM users&quot;) We can now execute such query with a specific page size using the :page_size option: {:ok, %Xandra.Page{} = page} = Xandra.execute(conn, prepared, [], page_size: 2) Since :page_size is 2, page will contain at most 2 rows: Enum.to_list(page) #=&gt; [%{&quot;first_name&quot; =&gt; &quot;Ross&quot;}, %{&quot;first_name&quot; =&gt; &quot;Rachel&quot;}] Now, we can pass page.paging_state as the value of the :paging_state option to let the paging start from where we left off: {:ok, %Xandra.Page{} = new_page} = Xandra.execute(conn, prepared, [], page_size: 2, paging_state: page.paging_state) Enum.to_list(page) #=&gt; [%{&quot;first_name&quot; =&gt; &quot;Joey&quot;}, %{&quot;first_name&quot; =&gt; &quot;Phoebe&quot;}] However, using :paging_state and :page_size directly with execute/4 is not recommended when the intent is to &quot;stream&quot; a query. For that, it&#39;s recommended to use stream_pages!/4. Also note that if the :paging_state option is set to nil, meaning there are no more pages to fetch, an ArgumentError exception will be raised; be sure to check for this with page.paging_state != nil."},{"ref":"Xandra.html#execute!/3","title":"Xandra.execute!/3","type":"function","doc":"Executes the given simple query, prepared query, or batch query, raising if there&#39;s an error. This function behaves exactly like execute/3, except that it returns successful results directly and raises on errors. Examples Xandra.execute!(conn, &quot;INSERT INTO users (name, age) VALUES (&#39;Jane&#39;, 29)&quot;) #=&gt; %Xandra.Void{}"},{"ref":"Xandra.html#execute!/4","title":"Xandra.execute!/4","type":"function","doc":"Executes the given simple query, prepared query, or batch query, raising if there&#39;s an error. This function behaves exactly like execute/4, except that it returns successful results directly and raises on errors. Examples statement = &quot;INSERT INTO users (name, age) VALUES (&#39;John&#39;, 43)&quot; Xandra.execute!(conn, statement, _params = [], consistency: :quorum) #=&gt; %Xandra.Void{}"},{"ref":"Xandra.html#prepare/3","title":"Xandra.prepare/3","type":"function","doc":"Prepares the given query. This function prepares the given statement on the Cassandra server. If preparation is successful and there are no network errors while talking to the server, {:ok, prepared} is returned, otherwise {:error, error} is returned. The returned prepared query can be run through execute/4, or used inside a batch (see Xandra.Batch). Errors returned by this function can be either Xandra.Error or Xandra.ConnectionError structs. See the module documentation for more information about errors. Supports all the options supported by DBConnection.prepare/3, and the following additional options: :force - (boolean) when true, forces the preparation of the query on the server instead of trying to read the prepared query from cache. See the &quot;Prepared queries cache&quot; section below. Defaults to false. :compressor - (module) the compressor module used to compress and decompress data. See the &quot;Compression&quot; section in the module documentation. By default, this option is not present. Prepared queries cache Since Cassandra prepares queries on a per-node basis (and not on a per-connection basis), Xandra internally caches prepared queries for each connection or pool of connections. This means that if you prepare a query that was already prepared, no action will be executed on the Cassandra server and the prepared query will be returned from the cache. If the Cassandra node goes down, however, the prepared query will be invalidated and trying to use the one from cache will result in a Xandra.Error. However, this is automatically handled by Xandra: when such an error is returned, Xandra will first retry to prepare the query and only return an error if the preparation fails. If you want to ensure a query is prepared on the server, you can set the :force option to true. Examples {:ok, prepared} = Xandra.prepare(conn, &quot;SELECT * FROM users WHERE id = ?&quot;) {:ok, _page} = Xandra.execute(conn, prepared, [_id = 1]) {:error, %Xandra.Error{reason: :invalid_syntax}} = Xandra.prepare(conn, &quot;bad syntax&quot;) # Force a query to be prepared on the server and not be read from cache: Xandra.prepare!(conn, &quot;SELECT * FROM users WHERE ID = ?&quot;, force: true)"},{"ref":"Xandra.html#prepare!/3","title":"Xandra.prepare!/3","type":"function","doc":"Prepares the given query, raising if there&#39;s an error. This function works exactly like prepare/3, except it returns the prepared query directly if preparation succeeds, otherwise raises the returned error. Examples prepared = Xandra.prepare!(conn, &quot;SELECT * FROM users WHERE id = ?&quot;) {:ok, _page} = Xandra.execute(conn, prepared, [_id = 1])"},{"ref":"Xandra.html#run/3","title":"Xandra.run/3","type":"function","doc":"Acquires a locked connection from conn and executes fun passing such connection as the argument. All options are forwarded to DBConnection.run/3 (and thus some of them to the underlying pool). The return value of this function is the return value of fun. Examples Preparing a query and executing it on the same connection: Xandra.run(conn, fn conn -&gt; prepared = Xandra.prepare!(conn, &quot;INSERT INTO users (name, age) VALUES (:name, :age)&quot;) Xandra.execute!(conn, prepared, %{&quot;name&quot; =&gt; &quot;John&quot;, &quot;age&quot; =&gt; 84}) end)"},{"ref":"Xandra.html#start_link/1","title":"Xandra.start_link/1","type":"function","doc":"Starts a new connection or pool of connections to Cassandra. This function starts a new connection or pool of connections to the provided Cassandra server. options is a list of both Xandra-specific options, as well as DBConnection options. Options These are the Xandra-specific options supported by this function: :nodes - (list of strings) the Cassandra nodes to connect to. Each node in the list has to be in the form &quot;ADDRESS:PORT&quot; or in the form &quot;ADDRESS&quot;: if the latter is used, the default port (9042) will be used for that node. Defaults to [&quot;127.0.0.1&quot;]. This option must contain only one node unless the :pool option is set to Xandra.Cluster; see the documentation for Xandra.Cluster for more information. :compressor - (module) the compressor module to use for compressing and decompressing data. See the &quot;Compression&quot; section in the module documentation. By default this option is not present. :authentication - (tuple) a two-element tuple: the authenticator module to use for authentication and its supported options. See the &quot;Authentication&quot; section in the module documentation. :atom_keys - (boolean) whether or not results of and parameters to execute/4 will have atom keys. If true, the result maps will have column names returned as atoms rather than as strings. Additionally, maps that represent named parameters will need atom keys. Defaults to false. The rest of the options are forwarded to DBConnection.start_link/2. For example, to start a pool of connections to Cassandra, the :pool option can be used: Xandra.start_link(pool: DBConnection.Poolboy) Note that this requires the poolboy dependency to be specified in your application. The following options have default values that are different from the default values provided by DBConnection: :idle_timeout - defaults to 30_000 (30 seconds) Examples # Start a connection: {:ok, conn} = Xandra.start_link() # Start a connection and register it under a name: {:ok, _conn} = Xandra.start_link(name: :xandra) # Start a named pool of connections: {:ok, _pool} = Xandra.start_link(name: :xandra_pool, pool: DBConnection.Poolboy) As the DBConnection documentation states, if using a pool it&#39;s necessary to pass a :pool option with the pool module being used to every call. For example: {:ok, _pool} = Xandra.start_link(name: :xandra_pool, pool: DBConnection.Poolboy) Xandra.execute!(:xandra_pool, &quot;SELECT * FROM users&quot;, _params = [], pool: DBConnection.Poolboy) Using a keyspace for new connections It is common to start a Xandra connection or pool of connections that will use a single keyspace for their whole life span. Doing something like: {:ok, conn} = Xandra.start_link() Xandra.execute!(conn, &quot;USE my_keyspace&quot;) will work just fine when you only have one connection. If you have a pool of connections (with the :pool option), however, the code above won&#39;t work: that code would start the pool, and then checkout one connection from the pool to execute the USE my_keyspace query. That specific connection will then be using the my_keyspace keyspace, but all other connections in the pool will not. Fortunately, DBConnection provides an option we can use to solve this problem: :after_connect. This option can specify a function that will be run after each single connection to Cassandra. This function will take a connection and can be used to setup that connection; since this function is run for every established connection, it will work well with pools as well. {:ok, conn} = Xandra.start_link(after_connect: fn(conn) -&gt; Xandra.execute(conn, &quot;USE my_keyspace&quot;) end) See the documentation for DBConnection.start_link/2 for more information about this option."},{"ref":"Xandra.html#stream_pages!/4","title":"Xandra.stream_pages!/4","type":"function","doc":"Streams the results of a simple query or a prepared query with the given params. This function can be used to stream the results of query so as not to load them entirely in memory. This function doesn&#39;t send any query to Cassandra right away: it will only execute queries as necessary when results are requested out of the returned stream. The returned value is a stream of Xandra.Page structs, where each of such structs contains at most as many rows as specified by the :page_size option. Every time an element is requested from the stream, query will be executed with params to get that result. In order to get each result from Cassandra, execute!/4 is used: this means that if there is an error (such as a network error) when executing the queries, that error will be raised. Simple or prepared queries Regardless of query being a simple query or a prepared query, this function will execute it every time a result is needed from the returned stream. For this reason, it is usually a good idea to use prepared queries when streaming. Options options supports all the options supported by execute/4, with the same default values. Examples prepared = Xandra.prepare!(conn, &quot;SELECT * FROM users&quot;) users_stream = Xandra.stream_pages!(conn, prepared, _params = [], page_size: 2) [%Xandra.Page{} = _page1, %Xandra.Page{} = _page2] = Enum.take(users_stream, 2)"},{"ref":"Xandra.html#t:conn/0","title":"Xandra.conn/0","type":"type","doc":""},{"ref":"Xandra.html#t:db_connection_start_option/0","title":"Xandra.db_connection_start_option/0","type":"type","doc":""},{"ref":"Xandra.html#t:error/0","title":"Xandra.error/0","type":"type","doc":""},{"ref":"Xandra.html#t:result/0","title":"Xandra.result/0","type":"type","doc":""},{"ref":"Xandra.html#t:start_option/0","title":"Xandra.start_option/0","type":"type","doc":""},{"ref":"Xandra.html#t:start_options/0","title":"Xandra.start_options/0","type":"type","doc":""},{"ref":"Xandra.html#t:statement/0","title":"Xandra.statement/0","type":"type","doc":""},{"ref":"Xandra.html#t:values/0","title":"Xandra.values/0","type":"type","doc":""},{"ref":"Xandra.html#t:xandra_start_option/0","title":"Xandra.xandra_start_option/0","type":"type","doc":""},{"ref":"Xandra.Authenticator.html","title":"Xandra.Authenticator","type":"behaviour","doc":"A behaviour module for implementing a Cassandra authenticator. Examples defmodule MyAuthenticator do @behaviour Xandra.Authenticator def response_body(options) do [&quot;user:&quot;, Keyword.fetch!(options, :user), &quot;_password:&quot;, Keyword.fetch!(options, :password)] end end To use the authenticator defined above: Xandra.start_link(authentication: {MyAuthenticator, user: &quot;foo&quot;, password: &quot;bar&quot;}) Xandra supports Cassandra&#39;s PasswordAuthenticator by default, see Xandra.Authenticator.Password for more information."},{"ref":"Xandra.Authenticator.html#c:response_body/1","title":"Xandra.Authenticator.response_body/1","type":"callback","doc":"Returns an iodata that&#39;s used as the response body to Cassandra&#39;s auth challenge."},{"ref":"Xandra.Authenticator.Password.html","title":"Xandra.Authenticator.Password","type":"module","doc":"A Xandra.Authenticator that implements support for PasswordAuthenticator to authenticate with Cassandra server. Example authenticator_options = [username: &quot;xandra&quot;, password: &quot;secret&quot;] Xandra.start_link(authentication: {Xandra.Authenticator.Password, authenticator_options})"},{"ref":"Xandra.Authenticator.Password.html#response_body/1","title":"Xandra.Authenticator.Password.response_body/1","type":"function","doc":"Returns an iodata that&#39;s used as the response body to Cassandra&#39;s auth challenge. Callback implementation for Xandra.Authenticator.response_body/1."},{"ref":"Xandra.Batch.html","title":"Xandra.Batch","type":"module","doc":"Represents a batch of simple and/or prepared queries. This module provides a data structure that can be used to group queries and execute them as a Cassandra BATCH query. Batch queries can be executed through Xandra.execute/3 and Xandra.execute!/3; see their respective documentation for more information."},{"ref":"Xandra.Batch.html#add/3","title":"Xandra.Batch.add/3","type":"function","doc":"Adds a query to the given batch. query has to be either a simple query (statement) or a prepared query. Note that parameters have to be added alongside their corresponding query when adding a query to a batch. In contrast with functions like Xandra.execute/4, simple queries in batch queries only support positional parameters and do not support named parameters; this is a current Cassandra limitation. If a map of named parameters is passed alongside a simple query, an ArgumentError exception is raised. Named parameters are supported with prepared queries. Examples prepared = Xandra.prepare!(conn, &quot;INSERT INTO users (name, age) VALUES (?, ?)&quot;) batch = Xandra.Batch.new() |&gt; Xandra.Batch.add(prepared, [&quot;Rick&quot;, 60]) |&gt; Xandra.Batch.add(prepared, [&quot;Morty&quot;, 14]) |&gt; Xandra.Batch.add(prepared, [&quot;Jerry&quot;, 35]) |&gt; Xandra.Batch.add(&quot;DELETE FROM users WHERE name = &#39;Jerry&#39;&quot;) Xandra.execute!(conn, batch)"},{"ref":"Xandra.Batch.html#new/1","title":"Xandra.Batch.new/1","type":"function","doc":"Creates a new batch query. type represents the type of the batch query (:logged, :unlogged, or :counter). See the Cassandra documentation for the meaning of these types. Examples batch = Xandra.Batch.new()"},{"ref":"Xandra.Batch.html#t:t/0","title":"Xandra.Batch.t/0","type":"type","doc":""},{"ref":"Xandra.Batch.html#t:t/1","title":"Xandra.Batch.t/1","type":"opaque","doc":""},{"ref":"Xandra.Batch.html#t:type/0","title":"Xandra.Batch.type/0","type":"type","doc":""},{"ref":"Xandra.Calendar.html","title":"Xandra.Calendar","type":"module","doc":""},{"ref":"Xandra.Calendar.html#date_from_unix_days/1","title":"Xandra.Calendar.date_from_unix_days/1","type":"function","doc":""},{"ref":"Xandra.Calendar.html#date_to_unix_days/1","title":"Xandra.Calendar.date_to_unix_days/1","type":"function","doc":""},{"ref":"Xandra.Calendar.html#time_from_nanoseconds/1","title":"Xandra.Calendar.time_from_nanoseconds/1","type":"function","doc":""},{"ref":"Xandra.Calendar.html#time_to_nanoseconds/1","title":"Xandra.Calendar.time_to_nanoseconds/1","type":"function","doc":""},{"ref":"Xandra.Cluster.html","title":"Xandra.Cluster","type":"module","doc":"A DBConnection.Pool pool that implements clustering support. This module implements a DBConnection.Pool pool that implements support for connecting to multiple nodes and executing queries on such nodes based on a given &quot;strategy&quot;. Usage To use this pool, the :pool option in Xandra.start_link/1 needs to be set to Xandra.Cluster. Xandra.Cluster is a &quot;proxy&quot; pool in the sense that it only proxies requests to other underlying pools of Xandra connections; the underlying pool can be specified with the :underlying_pool option. When you start a Xandra.Cluster connection, it will start one pool (:underlying_pool) of connections to each of the nodes specified in :nodes. The default :underlying_pool is DBConnection.Connection, which means by default only a single connection to each specified node will be established. Note that regardless of the underlying pool, Xandra.Cluster will establish one extra connection to each node in the specified list of nodes (used for internal purposes). Here is an example of how one could use Xandra.Cluster to connect to multiple nodes, while using :poolboy for pooling the connections to each node: Xandra.start_link([ nodes: [&quot;cassandra1.example.net&quot;, &quot;cassandra2.example.net&quot;], pool: Xandra.Cluster, underlying_pool: DBConnection.Poolboy, pool_size: 10, ]) The code above will establish a pool of ten connections to each of the nodes specified in :nodes, for a total of twenty connections going out of the current machine, plus two extra connections (one per node) used for internal purposes. Once a Xandra.Cluster pool is started, queries executed through such pool will be &quot;routed&quot; to nodes in the provided list of nodes; see the &quot;Load balancing strategies&quot; section below. Load balancing strategies For now, there are two load balancing &quot;strategies&quot; implemented: :random - it will choose one of the connected nodes at random and execute the query on that node. :priority - it will choose a node to execute the query according to the order nodes appear in :nodes. Disconnections and reconnections Xandra.Cluster also supports nodes disconnecting and reconnecting: if Xandra detects one of the nodes in :nodes going down, it will not execute queries against it anymore, but will start executing queries on it as soon as it detects such node is back up. If all specified nodes happen to be down when a query is executed, a Xandra.ConnectionError with reason {:cluster, :not_connected} will be returned. Options These are the options that Xandra.start_link/1 accepts when pool: Xandra.Cluster is passed to it: :underlying_pool - (module) the DBConnection.Pool pool used to pool connections to each of the specified nodes. :load_balancing - (atom) load balancing &quot;strategy&quot;. Defaults to :random. To pass options to the underlying pool, you can just pass them alongside other options to Xandra.start_link/1."},{"ref":"Xandra.Cluster.html#activate/4","title":"Xandra.Cluster.activate/4","type":"function","doc":""},{"ref":"Xandra.Cluster.html#checkin/3","title":"Xandra.Cluster.checkin/3","type":"function","doc":"Checkin a connection&#39;s state to the pool. The pool_ref is from the return of checkout/2. state is the lastest state of the connection. Callback implementation for DBConnection.Pool.checkin/3."},{"ref":"Xandra.Cluster.html#checkout/2","title":"Xandra.Cluster.checkout/2","type":"function","doc":"Checkout a connection&#39;s state from a pool. The returned pool_ref will be passed to checkin/3, disconnect/4 and stop/4. module and state are the module and state of the connection. Callback implementation for DBConnection.Pool.checkout/2."},{"ref":"Xandra.Cluster.html#child_spec/1","title":"Xandra.Cluster.child_spec/1","type":"function","doc":"Returns a specification to start this module under a supervisor. See Supervisor."},{"ref":"Xandra.Cluster.html#child_spec/3","title":"Xandra.Cluster.child_spec/3","type":"function","doc":"Create a supervisor child specification for the pool with module module, options opts and child specification options child_opts. Callback implementation for DBConnection.Pool.child_spec/3."},{"ref":"Xandra.Cluster.html#disconnect/4","title":"Xandra.Cluster.disconnect/4","type":"function","doc":"Checkin a connection&#39;s state to the pool and disconnect it with an exception. The pool_ref is from the return of checkout/2. state is the lastest state of the connection. Callback implementation for DBConnection.Pool.disconnect/4."},{"ref":"Xandra.Cluster.html#ensure_all_started/2","title":"Xandra.Cluster.ensure_all_started/2","type":"function","doc":"Ensure all applications necessary to run the pool are started. Callback implementation for DBConnection.Pool.ensure_all_started/2."},{"ref":"Xandra.Cluster.html#init/1","title":"Xandra.Cluster.init/1","type":"function","doc":"Invoked when the server is started. start_link/3 or start/3 will block until it returns. init_arg is the argument term (second argument) passed to start_link/3. Returning {:ok, state} will cause start_link/3 to return {:ok, pid} and the process to enter its loop. Returning {:ok, state, timeout} is similar to {:ok, state} except handle_info(:timeout, state) will be called after timeout milliseconds if no messages are received within the timeout. Returning {:ok, state, :hibernate} is similar to {:ok, state} except the process is hibernated before entering the loop. See c:handle_call/3 for more information on hibernation. Returning {:ok, state, {:continue, continue}} is similar to {:ok, state} except that immediately after entering the loop the c:handle_continue/2 callback will be invoked with the value continue as first argument. Returning :ignore will cause start_link/3 to return :ignore and the process will exit normally without entering the loop or calling c:terminate/2. If used when part of a supervision tree the parent supervisor will not fail to start nor immediately try to restart the GenServer. The remainder of the supervision tree will be started and so the GenServer should not be required by other processes. It can be started later with Supervisor.restart_child/2 as the child specification is saved in the parent supervisor. The main use cases for this are: The GenServer is disabled by configuration but might be enabled later. An error occurred and it will be handled by a different mechanism than the Supervisor. Likely this approach involves calling Supervisor.restart_child/2 after a delay to attempt a restart. Returning {:stop, reason} will cause start_link/3 to return {:error, reason} and the process to exit with reason reason without entering the loop or calling c:terminate/2. Callback implementation for GenServer.init/1."},{"ref":"Xandra.Cluster.html#start_link/2","title":"Xandra.Cluster.start_link/2","type":"function","doc":"Start and link to a pool of module connections with options opts. Callback implementation for DBConnection.Pool.start_link/2."},{"ref":"Xandra.Cluster.html#stop/4","title":"Xandra.Cluster.stop/4","type":"function","doc":"Stop a connection. The pool_ref is from the return of checkout/2. state is the lastest state of the connection. Callback implementation for DBConnection.Pool.stop/4."},{"ref":"Xandra.Cluster.html#update/2","title":"Xandra.Cluster.update/2","type":"function","doc":""},{"ref":"Xandra.Cluster.ControlConnection.html","title":"Xandra.Cluster.ControlConnection","type":"module","doc":""},{"ref":"Xandra.Cluster.ControlConnection.html#start_link/5","title":"Xandra.Cluster.ControlConnection.start_link/5","type":"function","doc":""},{"ref":"Xandra.Cluster.StatusChange.html","title":"Xandra.Cluster.StatusChange","type":"module","doc":""},{"ref":"Xandra.Compressor.html","title":"Xandra.Compressor","type":"behaviour","doc":"A behaviour to compress and decompress binary data. Modules implementing this behaviour can be used to compress and decompress data using one of the compression algorithms supported by Cassandra (as of now, lz4 and snappy). Example Let&#39;s imagine you implemented the snappy compression algorithm in your application: defmodule Snappy do def compress(binary), do: ... def decompress(binary), do: ... end You can then implement a module that implements the Xandra.Compressor behaviour and can be used to compress and decompress data flowing through the connection to Cassandra: defmodule SnappyXandraCompressor do @behaviour Xandra.Compressor def algorithm(), do: :snappy defdelegate compress(binary), to: Snappy defdelegate decompress(binary), to: Snappy end Now, this module can be passed as the value of the :compressor option to many functions in Xandra: Xandra.start_link(compressor: SnappyXandraCompressor) For more information on compression, see the &quot;Compression&quot; section in the documentation for Xandra."},{"ref":"Xandra.Compressor.html#c:algorithm/0","title":"Xandra.Compressor.algorithm/0","type":"callback","doc":"Specifies which algorithm this module will use to compress and decompress data."},{"ref":"Xandra.Compressor.html#c:compress/1","title":"Xandra.Compressor.compress/1","type":"callback","doc":"Compresses the given iodata according to the algorithm returned by algorithm/0."},{"ref":"Xandra.Compressor.html#c:decompress/1","title":"Xandra.Compressor.decompress/1","type":"callback","doc":"Deompresses the given binary according to the algorithm returned by algorithm/0."},{"ref":"Xandra.ConnectionError.html","title":"Xandra.ConnectionError","type":"exception","doc":"An exception struct that represents an error in the connection to the Cassandra server. For more information on when this error is returned or raised, see the documentation for the Xandra module. The :action field represents the action that was being performed when the connection error occurred. The :reason field represents the reason of the connection error: for network errors, this is usually a POSIX reason (like :econnrefused). The following Xandra-specific reasons are supported: {:unsupported_compression, algorithm} - this happens when a :compressor module has been specified in Xandra.start_link/1, but negotiating the connection algorithm fails because such compressor module uses an algorithm that the Cassandra server does not support. {:cluster, :not_connected} - this happens when a Xandra.Cluster-based connection is not connected to any node (for example, because all the specified nodes are currently down). See the documentation for Xandra.Cluster for more information. Since this struct is an exception, it is possible to raise it with Kernel.raise/1. If the intent is to format connection errors as strings (for example, for logging purposes), it is possible to use Exception.message/1 to get a formatted version of the error."},{"ref":"Xandra.ConnectionError.html#message/1","title":"Xandra.ConnectionError.message/1","type":"function","doc":"Callback implementation for Exception.message/1."},{"ref":"Xandra.ConnectionError.html#new/2","title":"Xandra.ConnectionError.new/2","type":"function","doc":""},{"ref":"Xandra.ConnectionError.html#t:t/0","title":"Xandra.ConnectionError.t/0","type":"type","doc":""},{"ref":"Xandra.Error.html","title":"Xandra.Error","type":"exception","doc":"An exception struct that represents an error returned by Cassandra. For more information on when this error is returned or raised, see the documentation for the Xandra module. The :reason field represents the reason (as an atom) of the error. For example, if the query you&#39;re trying to execute contains a syntax error, :reason will be :invalid_syntax. The :message field is a string that contains the exact error message that Cassandra returned. Since this struct is an exception, it is possible to raise it with Kernel.raise/1. If the intent is to format errors as strings (for example, for logging purposes), it is possible to use Exception.message/1 to get a formatted version of the error."},{"ref":"Xandra.Error.html#new/2","title":"Xandra.Error.new/2","type":"function","doc":""},{"ref":"Xandra.Error.html#t:t/0","title":"Xandra.Error.t/0","type":"type","doc":""},{"ref":"Xandra.Page.html","title":"Xandra.Page","type":"module","doc":"A struct that represents a page of rows. This struct represents a page of rows that have been returned by the Cassandra server in response to a query such as SELECT, but have not yet been parsed into Elixir values. This struct implements the Enumerable protocol and is therefore a stream. It is through this protocol that a Xandra.Page struct can be parsed into Elixir values. The simplest way of getting a list of single rows out of a Xandra.Page struct is to use something like Enum.to_list/1. Each element emitted when streaming out of a Xandra.Page struct is a map of string column names to their corresponding value. The following fields are public and can be accessed or relied on: paging_state - the current paging state. Its value can be used to check whether more pages are available to fetch after the given page. This is useful when implementing manual paging. See also the documentation for Xandra.execute/4. Examples statement = &quot;SELECT name, age FROM users&quot; %Xandra.Page{} = page = Xandra.execute!(conn, statement, _params = []) Enum.each(page, fn %{&quot;name&quot; =&gt; name, &quot;age&quot; =&gt; age} -&gt; IO.puts &quot;Read user with name \#{name} (age \#{age}) out of the database&quot; end)"},{"ref":"Xandra.Page.html#t:paging_state/0","title":"Xandra.Page.paging_state/0","type":"type","doc":""},{"ref":"Xandra.Page.html#t:t/0","title":"Xandra.Page.t/0","type":"opaque","doc":""},{"ref":"Xandra.Prepared.html","title":"Xandra.Prepared","type":"module","doc":"A data structure used to internally represent prepared queries."},{"ref":"Xandra.Prepared.html#t:t/0","title":"Xandra.Prepared.t/0","type":"opaque","doc":""},{"ref":"Xandra.RetryStrategy.html","title":"Xandra.RetryStrategy","type":"behaviour","doc":"A behaviour that handles how to retry failed queries. This behaviour makes it possible to customize the strategy that Xandra uses to retry failed queries. By default, Xandra does not retry failed queries, and does not provide any default retry strategy since retrying queries based on the failure reason is very tied to application logic. A module that implements the Xandra.RetryStrategy behaviour can be passed to several functions in the Xandra module: look at the documentation for Xandra for more information. Usage When a query fails and a retry strategy module was passed as an option, Xandra will: invoke the new/1 callback with the options passed to the failing function to initialize the given retry strategy ask the retry strategy whether to retry or error out (retry/3) until either the query succeeds or retry/3 says to error out The new/1 and retry/3 callbacks will be invoked in the same process that executed the original query. If retry/3 says to retry a query, such query will be retried on a different Xandra connection than the one the query was first executed through. For more information, see the documentation for retry/3. Examples This is an example of a retry strategy that retries a fixed number of times before failing. It reads the allowed number of retries from the options. defmodule MyApp.CounterRetryStrategy do @behaviour Xandra.RetryStrategy def new(options) do Keyword.fetch!(options, :retry_count) end def retry(_error, _options, _retries_left = 0) do :error end def retry(_error, options, retries_left) do {:retry, options, retries_left - 1} end end Another interesting example could be a retry strategy based on downgrading consistency: for example, we could execute all queries with a high consistency (such as :all) at first, and in case of failure, try again with a lower consistency (such as :quorum), finally giving up if that fails as well. defmodule MyApp.DowngradingConsistencyRetryStrategy do @behaviour Xandra.RetryStrategy def new(_options) do :no_state end def retry(_error, options, :no_state) do case Keyword.fetch(options, :consistency) do # No consistency was specified, so we don&#39;t bother to retry. :error -&gt; :error {:ok, :all} -&gt; {:retry, Keyword.put(options, :consistency, :quorum), :no_state} {:ok, _other} -&gt; :error end end end"},{"ref":"Xandra.RetryStrategy.html#c:new/1","title":"Xandra.RetryStrategy.new/1","type":"callback","doc":"Initializes the state of a retry strategy based on the given options."},{"ref":"Xandra.RetryStrategy.html#c:retry/3","title":"Xandra.RetryStrategy.retry/3","type":"callback","doc":"Determines whether to retry the failed query or return the error. The first argument is the error that caused the query to fail: for example, it could be a Xandra.Error struct with reason :read_timeout. This can be used to determine the retry strategy based on the failure reason. The second argument is the options given to the function that failed while executing the query. The third argument is the retry strategy state returned either by new/1 (if this was the first attempt to retry) or by subsequent calls to retry/3. If :error is returned, the function that was trying to execute the query will return the error to the caller instead of retrying. If {:retry, new_options, new_state} is returned, the function that was trying to execute the query will be invoked again with the same query and new_options as its options. new_state will be used if the query fails again: in that case, retry/3 will be invoked again with new_state as its third argument. This process will continue until either the query is executed successfully or this callback returns :error. Note that when {:retry, new_options, new_state} is returned, the query will be executed again on a different Xandra connection. This behaviour is particularly useful with pooled connections and especially when using Xandra.Cluster as the pool, since it will mean that there&#39;s a chance the retried query will be executed on a different node altogether."},{"ref":"Xandra.RetryStrategy.html#t:state/0","title":"Xandra.RetryStrategy.state/0","type":"type","doc":""},{"ref":"Xandra.SchemaChange.html","title":"Xandra.SchemaChange","type":"module","doc":"A struct that represents the result of a query that modifies the schema. This struct has the following fields: :effect - the type of change involved. It&#39;s one of &quot;CREATED&quot;, &quot;UPDATED&quot;, or &quot;DROPPED&quot;. :target - what has been modified. It&#39;s one of &quot;KEYSPACE&quot;, &quot;TABLE&quot;, or &quot;TYPE&quot;. :options - a map of options that depends on the value of :target: if target is &quot;KEYSPACE&quot;, the map will have the form %{keyspace: keyspace} if the target is &quot;TABLE&quot; or &quot;TYPE&quot;, the map will have the form %{keyspace: keyspace, subject: subject} where keyspace is the keyspace where the change happened and subject is the name of what changed (so the name of the changed table or type)"},{"ref":"Xandra.SchemaChange.html#t:t/0","title":"Xandra.SchemaChange.t/0","type":"type","doc":""},{"ref":"Xandra.SetKeyspace.html","title":"Xandra.SetKeyspace","type":"module","doc":"A struct that represents the result of a USE query. This struct only has the :keyspace field, which contains the keyspace that was set through the executed USE query."},{"ref":"Xandra.SetKeyspace.html#t:t/0","title":"Xandra.SetKeyspace.t/0","type":"type","doc":""},{"ref":"Xandra.Void.html","title":"Xandra.Void","type":"module","doc":"A struct that represents an empty Cassandra result. This struct is returned as the result of queries such as INSERT, UPDATE, or DELETE."},{"ref":"Xandra.Void.html#t:t/0","title":"Xandra.Void.t/0","type":"type","doc":""}]